import re
from pathlib import Path
from urllib.parse import urlsplit

from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.load import FATMAN_MANIFEST_FILENAME, load_manifest_from_yaml
from racetrack_client.log.logs import get_logger
from racetrack_client.utils.semver import SemanticVersion

logger = get_logger(__name__)


def load_validated_manifest(path: str) -> Manifest:
    """
    Load and validate manifest from a path. Raise exception in case of a defect.
    :param path path to a Fatman manifest file or to a directory with it
    :return loaded, valid Manifest
    """
    manifest_path = Path(path)
    if manifest_path.is_dir():
        manifest_path = manifest_path / FATMAN_MANIFEST_FILENAME

    with wrap_context('loading manifest'):
        manifest = load_manifest_from_yaml(manifest_path)
    with wrap_context('validating manifest'):
        validate_manifest(manifest)

    logger.debug(f'Manifest file "{manifest_path}" is valid')
    return manifest


def validate_manifest(manifest: Manifest):
    """Check whether manifest is valid. Raise exception in case of error"""
    if manifest.lang != 'python3':
        assert manifest.python is None, f'"python" configuration should be filled only for python jobs'
    if manifest.lang != 'golang':
        assert manifest.golang is None, f'"golang" configuration should be filled only for golang jobs'
    if manifest.lang != 'docker-http':
        assert manifest.docker is None, f'"docker" configuration should be filled only for docker-http jobs'

    assert re.match(r"[^@]+@[^@]+\.[^@]+", manifest.owner_email), '"owner_email" is not a valid email'

    with wrap_context('parsing Fatman version'):
        SemanticVersion(manifest.version)

    assert 1 <= manifest.replicas <= 15, 'replicas count out of allowed range'

    assert urlsplit(manifest.git.remote).scheme == 'https', 'git remote URL should be HTTPS'



from typing import Optional
from urllib.parse import urlsplit

from racetrack_client.client_config.client_config import ClientConfig


DEFAULT_LIFECYCLE_SCHEME = 'https'
DEFAULT_LIFECYCLE_PATH = '/ikp-rt/lifecycle'


def resolve_lifecycle_url(client_config: ClientConfig, lifecycle_name: Optional[str]) -> str:
    """
    Get full URL of Lifecycle based on given short name.
    If aliased name is found, resolve it to full URL it's pointing to.
    If the name was not provided, return default URL.
    """
    short_url = _resolve_short_lifecycle_url(client_config, lifecycle_name)
    return _infer_full_lifecycle_url(short_url)


def _resolve_short_lifecycle_url(client_config: ClientConfig, lifecycle_name: Optional[str]) -> str:
    if not lifecycle_name:
        return client_config.lifecycle_url
    if lifecycle_name in client_config.lifecycle_url_aliases:
        return client_config.lifecycle_url_aliases[lifecycle_name]
    return lifecycle_name


def _infer_full_lifecycle_url(url: str) -> str:
    """Apply default protocol and path if not provided."""
    split = urlsplit(url)
    if not split.scheme:
        url = f'{DEFAULT_LIFECYCLE_SCHEME}://{url}'
    split = urlsplit(url)  # parse again as path is recognized properly only with a scheme
    path = split.path if split.netloc else ''
    if not path:
        url = url + DEFAULT_LIFECYCLE_PATH
    return url



from urllib.parse import urlsplit


def trim_url(url: str) -> str:
    """Trim trailing slash from the end of URL address"""
    split = urlsplit(url)
    path = split.path
    while path.endswith('/'):
        path = path[:-1]
    split = split._replace(path=path)
    return split.geturl()



import dataclasses
from typing import Dict, List, Type, TypeVar

import dacite
import yaml

T = TypeVar("T")


def parse_dict_dataclass(obj_dict: Dict, clazz: Type[T]) -> T:
    """Cast dict object to expected dataclass"""
    return dacite.from_dict(data_class=clazz, data=obj_dict, config=dacite.Config(check_types=True, strict=True))


def parse_dict_dataclasses(obj_list: List[Dict], clazz: Type[T]) -> List[T]:
    """Cast list of dict objects to expected dataclass types"""
    return [parse_dict_dataclass(obj_dict, clazz) for obj_dict in obj_list]


def parse_yaml_dataclass(yaml_obj: str, clazz: Type[T]) -> T:
    """Parse YAML and convert it to expected dataclass"""
    data = yaml.load(yaml_obj, Loader=yaml.FullLoader)
    return parse_dict_dataclass(data, clazz)


def dataclass_to_yaml_str(dt) -> str:
    data_dict = dataclasses.asdict(dt)
    data_dict = _remove_none(data_dict)
    return yaml.dump(data_dict)


def _remove_none(obj):
    """Remove unwanted null values"""
    if isinstance(obj, list):
        return [_remove_none(x) for x in obj if x is not None]
    elif isinstance(obj, dict):
        return {k: _remove_none(v) for k, v in obj.items() if k is not None and v is not None}
    else:
        return obj



from typing import Optional, Union, Dict, List

import requests
from racetrack_client.log.context_error import ContextError


def parse_response(response: requests.Response, error_context: str) -> Optional[Union[Dict, List]]:
    """
    Ensure response was successful. If not, try to extract error message from it.
    :return: response parsed as JSON object

    This function handles all of these following cases:
    - if request went fine, return parsed JSON object,
    - if request went fine, but response is empty (eg. 204 status after DELETing), return None
    - if request went fine, but there was eg. JSON parsing error,
     raise "LC-API deploying error: Json is invalid in character 5"
    - if request completely failed with empty response (eg. due to Kubernilla issues),
    raise "LC-API deploying error: 404 Client Error: Not found for url http://localhost/blahblah"
    - if request failed gracefully from other RT component returning {"error": "you have no power here"},
    raise "LC-API deploying error: 500 Internal Server Error: you have no power here"
    """
    try:
        if 'application/json' in response.headers['content-type']:
            result = response.json()
        else:
            result = None

        if response.ok:
            return result

        if result is not None and 'error' in result:
            raise RuntimeError(f'{response.status_code} {response.reason}: {result.get("error")}')
        response.raise_for_status()
        return result
    except Exception as e:
        raise ResponseError(error_context, response.status_code) from e


class ResponseError(ContextError):
    def __init__(self, error_context: str, status_code: int):
        super().__init__(error_context)
        self.status_code = status_code



import dataclasses
from pathlib import Path
from typing import Optional, Dict
from urllib.parse import urlsplit
import tarfile
import io
from base64 import b64encode

import backoff
import requests

from racetrack_client.client.env import read_secret_vars, SecretVars
from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.client_config.client_config import ClientConfig, Credentials
from racetrack_client.client_config.io import load_client_config
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.utils.request import parse_response
from racetrack_client.log.logs import get_logger


logger = get_logger(__name__)

DEPLOYMENT_TIMEOUT_SECS = 60 * 60  # 1 hour


class DeploymentError(RuntimeError):
    def __init__(self, cause: BaseException):
        super().__init__()
        self.__cause__ = cause

    def __str__(self):
        return f'deployment error: {self.__cause__}'


def send_deploy_request(
    workdir: str,
    manifest: Optional[Manifest] = None,
    client_config: Optional[ClientConfig] = None,
    lifecycle_url: Optional[str] = None,
    force: bool = False,
):
    """
    Send request deploying a new Fatman to running Lifecycle instance
    :param workdir: directory with fatman.yaml manifest
    :param manifest: manifest to overwrite (if not taken from workdir)
    :param client_config: client configuration to use (if not the defaullt)
    :param lifecycle_url: URL to Racetrack server or alias name
    :param force: overwrite existing fatman without asking
    """
    if client_config is None:
        client_config = load_client_config()
    if manifest is None:
        manifest = load_validated_manifest(workdir)
        logger.debug(f'Manifest loaded: {manifest}')

    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    logger.info(f'deploying workspace "{workdir}" (job "{manifest.name}") to {lifecycle_url}')

    build_context = get_build_context(lifecycle_url, workdir)
    git_credentials = get_git_credentials(manifest, client_config)
    secret_vars = read_secret_vars(workdir, manifest)

    # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::DeployEndpoint` for server-side implementation
    r = requests.post(
        f'{lifecycle_url}/api/v1/deploy',
        json=_get_deploy_request_payload(manifest, git_credentials, secret_vars, build_context, force),
        headers=get_auth_request_headers(user_auth),
        verify=False,
    )
    response = parse_response(r, 'Lifecycle deploying error')
    deploy_id = response["id"]
    logger.info(f'job deployment requested: {deploy_id}')

    try:
        fatman_url = _wait_for_deployment_result(lifecycle_url, deploy_id, user_auth)
    except Exception as e:
        raise DeploymentError(e)

    logger.info(f'Fatman "{manifest.name}" has been deployed. Check out {fatman_url} to access your fatman')


def _get_deploy_request_payload(
        manifest: Manifest,
        git_credentials: Optional[Credentials],
        secret_vars: SecretVars,
        build_context: Optional[str],
        force: bool,
) -> Dict:
    return {
        "manifest": dataclasses.asdict(manifest),
        "git_credentials": dataclasses.asdict(git_credentials) if git_credentials is not None else None,
        "secret_vars": dataclasses.asdict(secret_vars),
        "build_context": build_context,
        "force": force,
    }


@backoff.on_exception(
    backoff.fibo, TimeoutError, max_value=5, max_time=DEPLOYMENT_TIMEOUT_SECS, jitter=None, logger=None
)
def _wait_for_deployment_result(lifecycle_url: str, deploy_id: str, user_auth: str) -> str:
    # see `lifecycle.endpoints.deploy::setup_deploy_endpoints::DeployIdEndpoint` for server-side implementation
    r = requests.get(
        f'{lifecycle_url}/api/v1/deploy/{deploy_id}',
        headers=get_auth_request_headers(user_auth),
        verify=False,
    )
    response = parse_response(r, 'Lifecycle deployment status')
    status = response['status'].lower()
    if status == 'failed':
        raise RuntimeError(response['error'])
    if status == 'done':
        return response.get('fatman', {}).get('pub_url')

    phase = response.get('phase')
    if phase:
        logger.info(f'deployment in progress: {phase}...')
    else:
        logger.info(f'deployment in progress...')
    raise TimeoutError('Deployment timeout error')


def get_git_credentials(manifest: Manifest, client_config: ClientConfig) -> Optional[Credentials]:
    if manifest.git.remote in client_config.git_credentials:
        return client_config.git_credentials[manifest.git.remote]

    split = urlsplit(manifest.git.remote)

    url = f'{split.scheme}://{split.netloc}{split.path}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.scheme}://{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    url = f'{split.netloc}'
    if url in client_config.git_credentials:
        return client_config.git_credentials[url]

    return None


def get_build_context(lifecycle_url: str, workdir: str) -> Optional[str]:
    """
    Avoid fetching from GitLab when using samples from local project repository.
    Send compressed build context instead in order to prevent the need to commit every change.
    """
    if not _is_url_localhost(lifecycle_url):
        return None

    logger.warning('transferring build context due to local development')
    return encode_build_context(workdir)


def _is_url_localhost(url: str) -> bool:
    split = urlsplit(url)
    return split.hostname in {'localhost', '127.0.0.1'}


def encode_build_context(workdir: str) -> str:
    """Compress build context directory to .tar.gz and encode it to base64 string"""
    tar_fileobj = io.BytesIO()
    workdir_path = Path(workdir)
    if workdir_path.is_file():
        workdir_path = workdir_path.parent
    with tarfile.open(fileobj=tar_fileobj, mode='w:gz') as tar:
        for p in workdir_path.iterdir():
            relative = p.relative_to(workdir_path)
            absolute = workdir_path / relative
            tar.add(str(absolute), arcname=str(relative))
    return b64encode(tar_fileobj.getvalue()).decode()



import requests
import logging

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.client_config import ClientConfig
from racetrack_client.utils.request import parse_response


class AuthError(RuntimeError):
    def __init__(self, cause: str):
        super().__init__()
        self.cause = cause

    def __str__(self):
        return f'authentication error: {self.cause}'


def set_user_auth(client_config: ClientConfig, lifecycle_url: str, user_auth: str):
    """
    You need to save the resulting config if you want to make changes persist
    """
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)

    if len(user_auth) == 0:
        if lifecycle_url in client_config.user_auths:
            del client_config.user_auths[lifecycle_url]
        else:
            raise RuntimeError(f'Missing {lifecycle_url} in Racetrack logged servers')
    else:
        client_config.user_auths[lifecycle_url] = user_auth


def get_user_auth(client_config: ClientConfig, lifecycle_url: str) -> str:
    if lifecycle_url in client_config.user_auths:
        return client_config.user_auths[lifecycle_url]

    if is_auth_required(lifecycle_url):
        raise AuthError(f"missing login. You need to do: racetrack login {lifecycle_url} <token>")

    return ''


def is_auth_required(lifecycle_url: str) -> bool:
    r = requests.get(
        f'{lifecycle_url}/api/v1/info',
        verify=False,
    )
    response = parse_response(r, 'Lifecycle response error')
    return response["auth_required"]



import dataclasses
from pathlib import Path
from typing import Dict, Optional

import dacite
import yaml
from racetrack_client.log.context_error import wrap_context
from racetrack_client.client_config.client_config import ClientConfig, Credentials
from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


def load_client_config() -> ClientConfig:
    """Load global options for a local client"""
    path = Path.home() / '.racetrack' / 'config.yaml'
    if not path.is_file():
        return ClientConfig()

    try:
        with path.open() as file:
            config_dict = yaml.load(file, Loader=yaml.FullLoader)
            config = dacite.from_dict(
                data_class=ClientConfig,
                data=config_dict,
                config=dacite.Config(cast=[]),
            )

            logger.info(f'client config loaded from {path}')
            return config
    except Exception as e:
        raise RuntimeError('loading client config failed') from e


def load_credentials_from_dict(credentials_dict: Optional[Dict]) -> Optional[Credentials]:
    if credentials_dict is None:
        return None

    with wrap_context('parsing credentials'):
        return dacite.from_dict(
            data_class=Credentials,
            data=credentials_dict,
            config=dacite.Config(cast=[]),
        )


def save_client_config(config: ClientConfig):
    data_dict = dataclasses.asdict(config)

    dir_path = Path.home() / '.racetrack'
    dir_path.mkdir(parents=True, exist_ok=True)
    config_path = dir_path / 'config.yaml'

    with open(config_path, 'w') as outfile:
        yaml.dump(data_dict, outfile)



from datetime import datetime, timezone

import pytz


def datetime_to_timestamp(dt: datetime) -> int:
    """Convert datetime.datetime to integer timestamp in seconds"""
    return int(dt.timestamp())


def timestamp_to_datetime(timestamp: int) -> datetime:
    """Convert integer timestamp in seconds to datetime.datetime"""
    return datetime.fromtimestamp(timestamp, tz=timezone.utc)


def datetime_to_str(dt: datetime) -> str:
    """Convert datetime to ISO 8601 format"""
    return dt.strftime('%Y-%m-%dT%H:%M:%S%z')


def now() -> datetime:
    """Return current datetime with UTC timezone set"""
    return datetime.now(tz=pytz.UTC)


def timestamp_pretty_ago(timestamp: int) -> str:
    """
    Convert past date to user-friendly description compared to current datetime.
    eg.: 'an hour ago', 'yesterday', '3 months ago', 'just now'
    """
    diff = now() - timestamp_to_datetime(timestamp)
    second_diff: int = diff.seconds
    day_diff: int = diff.days

    if day_diff < 0:
        return ''
    if day_diff == 0:
        if second_diff < 10:
            return "just now"
        if second_diff < 60:
            return f"{second_diff} seconds ago"
        if second_diff < 120:
            return "a minute ago"
        if second_diff < 3600:
            return f"{second_diff // 60} minutes ago"
        if second_diff < 7200:
            return "an hour ago"
        if second_diff < 86400:
            return f"{second_diff // 3600} hours ago"
    if day_diff == 1:
        return "yesterday"
    if day_diff < 7:
        return f"{day_diff} days ago"
    if day_diff // 7 == 1:
        return f"a week ago"
    if day_diff < 31:
        return f"{day_diff // 7} weeks ago"
    if day_diff // 30 == 1:
        return f"a month ago"
    if day_diff < 365:
        return f"{day_diff // 30} months ago"
    if day_diff // 365 == 1:
        return f"a year ago"
    return f"{day_diff // 365} years ago"



import io
import subprocess
import sys
import threading
from pathlib import Path
from typing import Optional, Callable

from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


def shell(cmd: str, workdir: Optional[Path] = None):
    """
    Run system shell command.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    :raises:
        CommandError: in case of non-zero command exit code.
    """
    _run_shell_command(cmd, workdir)


def shell_output(cmd: str, workdir: Optional[Path] = None, print_stdout: bool = False) -> str:
    """
    Run system shell command and return its output.
    Print live stdout as it comes (line by line) and capture entire output in case of errors.
    """
    captured_stream = _run_shell_command(cmd, workdir, print_stdout)
    return captured_stream.getvalue()


def _run_shell_command(cmd: str, workdir: Optional[Path] = None, print_stdout: bool = True) -> io.StringIO:
    logger.debug(f'Command: {cmd}')
    if len(cmd) > 4096:  # see https://github.com/torvalds/linux/blob/v5.11/drivers/tty/n_tty.c#L1681
        raise RuntimeError('maximum tty line length has been exceeded')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=workdir)
    try:
        # fork command output to stdout & captured buffer
        captured_stream = io.StringIO()
        for line in iter(process.stdout.readline, b''):
            line_str = line.decode()
            if print_stdout:
                sys.stdout.write(line_str)
                sys.stdout.flush()
            captured_stream.write(line_str)

        process.wait()
        if process.returncode != 0:
            stdout = captured_stream.getvalue()
            raise CommandError(cmd, stdout, process.returncode)
        return captured_stream
    except KeyboardInterrupt:
        logger.warning('killing subprocess')
        process.kill()
        raise


class CommandOutputStream:
    def __init__(self,
                 cmd: str,
                 on_next_line: Callable[[str], None],
                 on_error: Callable[['CommandError'], None] = None,
                 workdir: Optional[Path] = None,
                 print_stdout: bool = False):
        """
        Manager of a shell command process running in background,
        monitoring stdout lines coming from it and streaming them.
        :param cmd: shell command
        :param on_next_line: callback function called whenever the new stdout line appears
        :param on_error: callback function called when process ends with non-zero return code
        :param workdir: working directory to call command with
        :param print_stdout: whether to write output also to stdout
        """
        self.stop = False
        logger.debug(f'Command: {cmd}')
        self.process = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True, cwd=workdir)

        def monitor_output(stream):
            captured_stream = io.StringIO()
            for line in iter(stream.process.stdout.readline, b''):
                if stream.stop:
                    break
                line_str = line.decode()
                if print_stdout:
                    sys.stdout.write(line_str)
                    sys.stdout.flush()
                on_next_line(line_str)
                captured_stream.write(line_str)

            stream.process.wait()
            if stream.process.returncode != 0 and on_error is not None:
                stdout = captured_stream.getvalue()
                on_error(CommandError(cmd, stdout, stream.process.returncode))

        self.thread = threading.Thread(
            target=monitor_output,
            args=(self,),
            daemon=True,
        )
        self.thread.start()

    def interrupt(self):
        self.stop = True
        self.process.terminate()


class CommandError(RuntimeError):
    def __init__(self, cmd: str, stdout: str, returncode: int):
        super().__init__()
        self.cmd = cmd
        self.stdout = stdout
        self.returncode = returncode

    def __str__(self):
        return f'command failed: {self.cmd}: {self.stdout}'



import argparse
import sys
from racetrack_client.client.delete import delete_fatman

import yaml
import urllib3

from racetrack_client import __version__
from racetrack_client.client.deploy import send_deploy_request, DeploymentError
from racetrack_client.client.logs import show_runtime_logs, show_build_logs
from racetrack_client.client_config.auth import set_user_auth, AuthError
from racetrack_client.client_config.client_config import ClientConfig
from racetrack_client.client_config.io import load_client_config, save_client_config
from racetrack_client.client_config.update import set_credentials, set_config_setting, set_config_url_alias
from racetrack_client.log.exception import log_exception
from racetrack_client.log.logs import configure_logs
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest.validate import load_validated_manifest

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)

logger = get_logger(__name__)


def main():
    parser = argparse.ArgumentParser(description='CLI client tool for deploying workloads to Racetrack')
    parser.add_argument('-v', '--verbose', action='count', default=0, help='enable verbose mode')
    subparsers = parser.add_subparsers()

    def _print_help(_: argparse.Namespace):
        parser.print_help(sys.stderr)

    parser.set_defaults(func=_print_help)

    # racetrack deploy
    parser_deploy = subparsers.add_parser(
        'deploy', help='Send request deploying a Fatman to the Racetrack cluster')
    parser_deploy.add_argument('workdir', default='.', nargs='?', help='directory with fatman.yaml manifest')
    parser_deploy.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_deploy.add_argument('--force', action='store_true', help='overwrite existing fatman')
    parser_deploy.set_defaults(func=_deploy)

    # racetrack validate
    parser_validate = subparsers.add_parser('validate', help='Validate Fatman manifest file')
    parser_validate.add_argument('path', default='.', nargs='?',
                                 help='path to a Fatman manifest file or to a directory with it')
    parser_validate.set_defaults(func=_validate)

    # racetrack logs
    parser_logs = subparsers.add_parser('logs', help='Show logs from fatman output')
    parser_logs.add_argument('workdir', default='.', nargs='?', help='directory with fatman.yaml manifest')
    parser_logs.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_logs.add_argument('--tail', default=20, nargs='?', type=int, help='number of recent lines to show')
    parser_logs.add_argument('--follow', '-f', action='store_true', help='follow logs output stream')
    parser_logs.set_defaults(func=_logs)

    # racetrack build-logs
    parser_build_logs = subparsers.add_parser('build-logs', help='Show build logs from fatman image building')
    parser_build_logs.add_argument('workdir', default='.', nargs='?', help='directory with fatman.yaml manifest')
    parser_build_logs.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_build_logs.add_argument('--tail', default=0, nargs='?', type=int,
                                   help='number of recent lines to show, all logs by default')
    parser_build_logs.set_defaults(func=_build_logs)

    # racetrack delete
    parser_delete = subparsers.add_parser('delete', help='Delete fatman instance')
    parser_delete.add_argument('workdir', default='.', nargs='?', help='directory with fatman.yaml manifest')
    parser_delete.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_delete.add_argument('--version', nargs='?', type=str,
                               help='fatman version to delete')
    parser_delete.set_defaults(func=_delete_fatman)

    # racetrack version
    parser_version = subparsers.add_parser('version', help='Show the version information')
    parser_version.set_defaults(func=_version)

    # racetrack config
    parser_config = subparsers.add_parser('config', help='Set local options for a Racetrack client')
    subparsers_config = parser_config.add_subparsers()

    # racetrack config show
    parser_config_show = subparsers_config.add_parser('show', help='Show racetrack config values')
    parser_config_show.set_defaults(func=_show_config)

    # racetrack config racetrack_url
    parser_config_racetrack_url = subparsers_config.add_parser(
        'racetrack_url', help='Set default Racetrack URL address')
    parser_config_racetrack_url.add_argument('setting_value', help='setting value')
    parser_config_racetrack_url.set_defaults(func=_set_config_racetrack_url)

    # racetrack config credentials
    parser_config_credentials = subparsers_config.add_parser(
        'credentials', help='Manage credentials for git repository access')
    subparsers_config_credentials = parser_config_credentials.add_subparsers()

    # racetrack config credentials set
    parser_config_credentials_set = subparsers_config_credentials.add_parser(
        'set', help='Set credentials for reading git repository')
    parser_config_credentials_set.add_argument('repo_url', help='git remote URL')
    parser_config_credentials_set.add_argument('username', help='username for git authentication')
    parser_config_credentials_set.add_argument('token_password', help='password or token for git authentication')
    parser_config_credentials_set.set_defaults(func=_set_config_credentials)

    # racetrack config alias
    parser_config_alias = subparsers_config.add_parser('alias', help='Manage aliases for Racetrack server URLs')
    subparsers_config_alias = parser_config_alias.add_subparsers()

    # racetrack config alias set
    parser_config_alias_set = subparsers_config_alias.add_parser(
        'set', help='Set up an alias for Racetrack server URL')
    parser_config_alias_set.add_argument('alias', help='short name for an environment')
    parser_config_alias_set.add_argument('racetrack_url', help='Racetrack server URL address')
    parser_config_alias_set.set_defaults(func=_set_config_url_alias)

    # racetrack login
    parser_login = subparsers.add_parser('login', help='Save user token for Racetrack server')
    parser_login.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_login.add_argument('user_token', default='', nargs='?', help='User token from RT user profile')
    parser_login.set_defaults(func=_login)

    # racetrack logout
    parser_logout = subparsers.add_parser('logout', help='Remove user token for Racetrack server')
    parser_logout.add_argument('racetrack_url', default='', nargs='?', help='URL to Racetrack server or alias name')
    parser_logout.set_defaults(func=_logout)

    args: argparse.Namespace = parser.parse_args()

    try:
        configure_logs(verbosity=args.verbose)
        args.func(args)
    except (DeploymentError, AuthError) as e:
        logger.error(str(e))  # no need for client's stacktrace in case of well known errors
    except Exception as e:
        log_exception(e)


def _deploy(args: argparse.Namespace):
    send_deploy_request(args.workdir, lifecycle_url=args.racetrack_url, force=args.force)


def _validate(args: argparse.Namespace):
    load_validated_manifest(args.path)


def _set_config_credentials(args: argparse.Namespace):
    set_credentials(args.repo_url, args.username, args.token_password)


def _set_config_racetrack_url(args: argparse.Namespace):
    set_config_setting('racetrack_url', args.setting_value)


def _set_config_url_alias(args: argparse.Namespace):
    set_config_url_alias(args.alias, args.racetrack_url)


def _logs(args: argparse.Namespace):
    show_runtime_logs(args.workdir, args.racetrack_url, args.tail, args.follow)


def _build_logs(args: argparse.Namespace):
    show_build_logs(args.workdir, args.racetrack_url, args.tail)


def _delete_fatman(args: argparse.Namespace):
    delete_fatman(args.workdir, args.racetrack_url, args.version)


def _show_config(args: argparse.Namespace):
    client_config = load_client_config()
    print(yaml.dump(client_config))


def _version(_: argparse.Namespace):
    print(f'racetrack-client version {__version__}')


def _login(args: argparse.Namespace):
    client_config: ClientConfig = load_client_config()
    set_user_auth(client_config, args.racetrack_url, args.user_token)
    save_client_config(client_config)
    logger.info(f'Logged to Racetrack: {args.racetrack_url}')


def _logout(args: argparse.Namespace):
    client_config: ClientConfig = load_client_config()
    set_user_auth(client_config, args.racetrack_url, "")
    save_client_config(client_config)
    logger.info(f'Logged out from Racetrack: {args.racetrack_url}')



from dataclasses import dataclass
from typing import Optional, List, Dict, Any


@dataclass
class Manifest:
    """Fatman Manifest file - build recipe to get deployable image from source code workspace"""

    # name of the Fatman Workload
    name: str

    # email address of the Fatman's owner to reach out
    owner_email: str

    git: 'GitManifest'

    # version of the Fatman
    version: str = '0.0.1'

    # Language wrapper used to embed model
    lang: str = 'python3'

    # Python-specific configuration
    python: Optional['PythonManifest'] = None

    # Go/Golang-specific configuration
    golang: Optional['GoManifest'] = None

    # Docker-specific configuration
    docker: Optional['DockerManifest'] = None

    # type of deployed image: docker image, packer, AMI
    image_type: str = 'docker'

    # system-wide packages that should be installed with apt
    system_dependencies: Optional[List[str]] = None

    # env vars for building
    build_env: Optional[Dict[str, str]] = None
    # env vars for runtime
    runtime_env: Optional[Dict[str, str]] = None
    # secret env vars loaded from an external file applied on building
    secret_build_env_file: Optional[str] = None
    # secret env vars loaded from an external file applied at runtime
    secret_runtime_env_file: Optional[str] = None

    # labels - fatman metadata for humans
    labels: Optional[Dict[str, Any]] = None

    # list of public fatman endpoints that can be accessed without authentication
    public_endpoints: Optional[List[str]] = None

    # number of instances of the Fatman
    replicas: int = 1


@dataclass
class PythonManifest:
    requirements_path: Optional[str] = None
    # path to a Python file with a entrypoint class
    entrypoint_path: str = ''
    # base name of Python entrypoint class
    entrypoint_class: str = ''


@dataclass
class GoManifest:
    # relative path to Go modules requirements
    gomod: str = 'go.mod'


@dataclass
class GitManifest:
    # URL of git remote: HTTPS, SSH or directory path to a remote repository
    remote: str
    branch: Optional[str] = None
    # subdirectory relative to git repo root
    directory: str = '.'


@dataclass
class DockerManifest:
    dockerfile_path: Optional[str] = None



from configparser import ConfigParser
from dataclasses import dataclass, field
from pathlib import Path
from typing import Dict, Optional

from racetrack_client.log.context_error import wrap_context
from racetrack_client.manifest import Manifest
from racetrack_client.utils.dataclass import parse_dict_dataclass
from racetrack_client.log.logs import get_logger

logger = get_logger(__name__)


@dataclass
class SecretVars:
    build_env: Dict[str, str] = field(default_factory=dict)
    runtime_env: Dict[str, str] = field(default_factory=dict)


def read_secret_vars(workdir: str, manifest: Manifest) -> SecretVars:
    secrets = SecretVars()
    secrets.build_env = read_secret_vars_from_file(workdir, manifest.secret_build_env_file, 'secret build vars')
    secrets.runtime_env = read_secret_vars_from_file(workdir, manifest.secret_runtime_env_file, 'secret runtime vars')
    return secrets


def read_secret_vars_from_file(workdir: str, secret_env_file: Optional[str], vars_name: str) -> Dict:
    if not secret_env_file:
        return {}

    secret_env_path = Path(workdir) / secret_env_file
    assert secret_env_path.is_file(), f"secret env file doesn't exist: {secret_env_path}"
    logger.info(f'reading {vars_name} from {secret_env_path}')
    return read_env_vars_from_file(secret_env_path)


def read_env_vars_from_file(path: Path) -> Dict[str, str]:
    with wrap_context('reading vars from env file'):
        config = ConfigParser()
        config.optionxform = str
        config.read_string("[config]\n" + path.read_text())
        env_dict = dict()
        for k, v in config["config"].items():
            env_dict[k] = v.strip('"').strip("'")
        return env_dict


def load_secret_vars_from_dict(secrets_dict: Optional[Dict[str, str]]) -> SecretVars:
    if secrets_dict is None:
        return SecretVars()
    with wrap_context('parsing secret_vars'):
        return parse_dict_dataclass(secrets_dict, SecretVars)


def merge_env_vars(env1: Optional[Dict[str, str]], env2: Optional[Dict[str, str]]) -> Dict[str, str]:
    if env1 is None:
        env1 = {}
    if env2 is None:
        env2 = {}
    return {**env1, **env2}


def hide_env_vars(runtime_env: Dict[str, str], build_env: Dict[str, str]) -> Dict[str, str]:
    """Shadow build vars in runtime env by adding empty values for them."""
    shadow = {k: '' for k, v in build_env.items()}
    return {**shadow, **runtime_env}



from typing import Optional

import requests

from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.utils.request import parse_response


logger = get_logger(__name__)


def delete_fatman(workdir: str, lifecycle_url: Optional[str], version: Optional[str]):
    """Delete versioned fatman instance"""
    client_config = load_client_config()
    manifest = load_validated_manifest(workdir)
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)

    version = version or manifest.version

    logger.debug(f'Deleting fatman "{manifest.name}" v{version} from {lifecycle_url}...')
    r = requests.delete(
        f'{lifecycle_url}/api/v1/fatman/{manifest.name}/{version}',
        headers=get_auth_request_headers(user_auth),
        verify=False,
    )
    parse_response(r, 'Lifecycle response error')
    logger.info(f'Fatman "{manifest.name}" v{version} has been deleted from {lifecycle_url}')



from typing import Optional

import requests

from racetrack_client.utils.auth import get_auth_request_headers
from racetrack_client.client.socketio import LogsConsumer
from racetrack_client.client_config.alias import resolve_lifecycle_url
from racetrack_client.client_config.auth import get_user_auth
from racetrack_client.client_config.io import load_client_config
from racetrack_client.log.logs import get_logger
from racetrack_client.manifest import Manifest
from racetrack_client.manifest.validate import load_validated_manifest
from racetrack_client.utils.request import parse_response


logger = get_logger(__name__)


def show_runtime_logs(workdir: str, lifecycle_url: Optional[str], tail: int, follow: bool):
    """
    Show logs from fatman output
    :param workdir: directory with fatman.yaml manifest
    :param lifecycle_url: URL to Lifecycle API
    :param tail: number of recent lines to show
    :param follow: whether to follow logs output stream
    """
    client_config = load_client_config()
    manifest = load_validated_manifest(workdir)
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)
    logger.info(f'Retrieving runtime logs of fatman "{manifest.name}" v{manifest.version} from {lifecycle_url}...')

    if follow:
        _show_runtime_logs_following(lifecycle_url, manifest, tail)
    else:
        _show_runtime_logs_once(lifecycle_url, manifest, tail, user_auth)


def _show_runtime_logs_once(lifecycle_url: str, manifest: Manifest, tail: int, user_auth: str):
    r = requests.get(
        f'{lifecycle_url}/api/v1/fatman/{manifest.name}/{manifest.version}/logs',
        params={'tail': tail},
        headers=get_auth_request_headers(user_auth),
        verify=False,
    )
    response = parse_response(r, 'Lifecycle response error')
    logs: str = response['logs']
    log_lines = len(logs.splitlines())
    logger.info(f'Viewing the latest logs from fatman "{manifest.name}" v{manifest.version} below'
                f' ({log_lines} lines):\n---')
    print(logs)


def _show_runtime_logs_following(lifecycle_url: str, manifest: Manifest, tail: int):
    def on_next_line(line: str):
        print(line.strip())

    resource_properties = {
        'fatman_name': manifest.name,
        'fatman_version': manifest.version,
        'tail': tail,
    }
    consumer = LogsConsumer(lifecycle_url, 'ikp-rt/lifecycle/socket.io', resource_properties, on_next_line)
    logger.info(f'Streaming live logs from fatman "{manifest.name}" v{manifest.version} below:\n---')
    consumer.connect_and_wait()


def show_build_logs(workdir: str, lifecycle_url: Optional[str], tail: int = 0):
    """
    Show output of latest fatman image building process
    :param workdir: directory with fatman.yaml manifest
    :param lifecycle_url: URL to Lifecycle API
    :param tail: number of recent lines to show. If zero, all logs are displayed.
    """
    client_config = load_client_config()
    manifest = load_validated_manifest(workdir)
    lifecycle_url = resolve_lifecycle_url(client_config, lifecycle_url)
    user_auth = get_user_auth(client_config, lifecycle_url)
    logger.info(f'Retrieving build logs of fatman "{manifest.name}" v{manifest.version} from {lifecycle_url}...')

    r = requests.get(
        f'{lifecycle_url}/api/v1/fatman/{manifest.name}/{manifest.version}/build-logs',
        params={'tail': tail},
        headers=get_auth_request_headers(user_auth),
        verify=False,
    )
    response = parse_response(r, 'Lifecycle response error')
    logs: str = response['logs']
    log_lines = len(logs.splitlines())
    logger.info(f'Recent build logs of "{manifest.name}" fatman ({log_lines} lines):\n')
    print(logs)



from enum import Enum
from typing import Dict

RT_USER_AUTH_HEADER = "X-Racetrack-User-Auth"
RT_INTERNAL_AUTH_HEADER = "X-Racetrack-Internal-Auth"
RT_FATMAN_AUTH_HEADER = "X-Racetrack-Fatman-Auth"
RT_ESC_AUTH_HEADER = 'X-Racetrack-Esc-Auth'

RT_SESSION_USER_AUTH_KEY = 'RACETRACK_SESSION_USER_AUTH'


class RacetrackAuthMethod(Enum):
    USER = 'racetrackUserAuth'
    INTERNAL = 'racetrackInternalAuth'
    FATMAN = 'racetrackFatmanAuth'
    ESC = 'racetrackESCAuth'


def get_auth_request_headers(user_auth: str) -> Dict:
    return {
        RT_USER_AUTH_HEADER: user_auth if user_auth != "" else None
    }



from multiprocessing.sharedctypes import Value
import re
from typing import Any, Callable, List, Optional, TypeVar


T = TypeVar("T")


class SemanticVersion:
    """
    Version numbering adhering to Semantic Versioning Specification (SemVer)
    It should match `MAJOR.MINOR.PATCH[LABEL]` format, eg.:
    - 1.2.3
    - 10.0.1-alpha
    - 0.0.0-dev
    """

    version_pattern = re.compile(r'(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?P<label>.*)')

    def __init__ (self, vstring):
        match = self.version_pattern.fullmatch(vstring)
        if not match:
            raise ValueError(f"Version '{vstring}' doesn't match SemVer format 'X.Y.Z[-label]'")
        self.major = int(match.group('major'))
        self.minor = int(match.group('minor'))
        self.patch = int(match.group('patch'))
        self.label = match.group('label')

    @staticmethod
    def find_latest_stable(objects: List[T], key: Callable[[T], str]) -> Optional[T]:
        """
        Find object with the latest stable version - highest version disregarding versions with labels, eg "-dev"
        :param objects: List of objects to compare
        :param key: Callable to extract version from an object
        :return: Object from a list having the latest stable version
        """
        versions_objects = {SemanticVersion(key(obj)): obj for obj in objects}
        # filter stable versions
        versions_objects = {version: obj for version, obj in versions_objects.items()
                            if not version.label}
        if not versions_objects:
            return None

        latest_version = max(versions_objects.keys())
        return versions_objects[latest_version]

    @staticmethod
    def find_latest_wildcard(pattern: 'SemanticVersionPattern', objects: List[T], key: Callable[[T], str]) -> Optional[T]:
        """
        Find object with the latest (highest), stable (non-dev) version matching pattern
        :param pattern: version pattern containing "x" wildcards, eg. "1.2.x", "2.x"
        :param objects: List of objects to compare
        :param key: Callable to extract version from an object
        :return: Object from a list having the latest version matching pattern
        """
        versions_objects = {SemanticVersion(key(obj)): obj for obj in objects}
        # filter stable versions matching pattern
        versions_objects = {version: obj for version, obj in versions_objects.items()
                            if not version.label and pattern.matches(version)}
        if not versions_objects:
            return None

        latest_version = max(versions_objects.keys())
        return versions_objects[latest_version]


    def __str__ (self):
        return f'{self.major}.{self.minor}.{self.patch}{self.label}'

    def _compare(self, other) -> int:
        """
        Returns zero if objects are considered "equal",
        negative if self is less than other, otherwise return positive.
        """
        if isinstance(other, str):
            other = SemanticVersion(other)

        if self.major != other.major:
            return self.major - other.major
        if self.minor != other.minor:
            return self.minor - other.minor
        if self.patch != other.patch:
            return self.patch - other.patch

        return self._compare_labels(self.label, other.label)

    def __eq__(self, other):
        return self._compare(other) == 0

    def __lt__(self, other):
        return self._compare(other) < 0

    def __le__(self, other):
        return self._compare(other) <= 0

    def __gt__(self, other):
        return self._compare(other) > 0

    def __ge__(self, other):
        return self._compare(other) >= 0

    @staticmethod
    def _compare_labels(label1: str, label2: str) -> int:
        if label1 == label2:
            return 0
        # According to Semver, pre-release version has lower precedence than a normal version
        # Example: 1.0.0-alpha < 1.0.0.
        if not label2:
            return -1
        if not label1:
            return 1

        if label1 < label2:
            return -1
        else:
            return 1
    
    def __members(self):
        return (self.major, self.minor, self.patch, self.label)

    def __eq__(self, other):
        if type(other) is type(self):
            return self.__members() == other.__members()
        else:
            return False

    def __hash__(self):
        return hash(self.__members())


class SemanticVersionPattern:
    valid_pattern_regex = re.compile(r'(\d+|x)(\.(\d+|x))*')  # 1.2.x or 1.x or 1.x.x

    def __init__(self, pattern: str) -> None:
        """
        :param pattern: version pattern containing "x" wildcards, eg. "1.2.x", "2.x"
        """
        match = self.valid_pattern_regex.fullmatch(pattern)
        if not match:
            raise ValueError(f"Version pattern '{pattern}' is invalid, should be '1.x' or '1.2.x'")
        if 'x' not in pattern:
            raise ValueError(f"Version pattern '{pattern}' doesn't contain 'x' wildcard")

        self.regex_pattern = re.compile(pattern.replace('.', r'\.').replace('x', r'.+'))

    def matches(self, version: SemanticVersion) -> bool:
        return self.regex_pattern.fullmatch(str(version)) is not None

    @staticmethod
    def is_wildcard_pattern(pattern: str) -> bool:
        match = SemanticVersionPattern.valid_pattern_regex.fullmatch(pattern)
        if not match:
            return False
        return 'x' in pattern
